@@ -1,5 +1,6 @@
 // There seem to be false positives with pyo3
 #![allow(clippy::borrow_deref_ref)]
+use ::biscuit_auth::RootKeyProvider;
 use chrono::DateTime;
 use chrono::TimeZone;
 use chrono::Utc;
@@ -41,15 +42,54 @@ create_exception!(
     pyo3::exceptions::PyException
 );
 
+struct PyKeyProvider {
+    py_value: PyObject,
+}
+
+impl RootKeyProvider for PyKeyProvider {
+    fn choose(&self, kid: Option<u32>) -> Result<PublicKey, error::Format> {
+        Python::with_gil(|py| {
+            if self.py_value.as_ref(py).is_callable() {
+                let result = self
+                    .py_value
+                    .call1(py, (kid,))
+                    .map_err(|_| error::Format::UnknownPublicKey)?;
+                let py_pk: PyPublicKey = result
+                    .extract(py)
+                    .map_err(|_| error::Format::UnknownPublicKey)?;
+                Ok(py_pk.0)
+            } else {
+                let py_pk: PyPublicKey = self
+                    .py_value
+                    .extract(py)
+                    .map_err(|_| error::Format::UnknownPublicKey)?;
+                Ok(py_pk.0)
+            }
+        })
+    }
+}
+
 /// Builder class allowing to create a biscuit from a datalog block
 ///
 /// :param source: a datalog snippet
+/// :type source: str, optional
+/// :param parameters: values for the parameters in the datalog snippet
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "BiscuitBuilder")]
 pub struct PyBiscuitBuilder(builder::BiscuitBuilder);
 
 #[pymethods]
 impl PyBiscuitBuilder {
     /// Create a builder from a datalog snippet and optional parameter values
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
     #[new]
     fn new(
         source: Option<String>,
@@ -58,11 +98,7 @@ impl PyBiscuitBuilder {
     ) -> PyResult<PyBiscuitBuilder> {
         let mut builder = PyBiscuitBuilder(builder::BiscuitBuilder::new());
         if let Some(source) = source {
-            builder.add_code_with_parameters(
-                &source,
-                parameters.unwrap_or_default(),
-                scope_parameters.unwrap_or_default(),
-            )?;
+            builder.add_code(&source, parameters, scope_parameters)?;
         }
         Ok(builder)
     }
@@ -70,7 +106,9 @@ impl PyBiscuitBuilder {
     /// Build a biscuit token, using the provided private key to sign the authority block
     ///
     /// :param root: a keypair that will be used to sign the authority block
+    /// :type root: PrivateKey
     /// :return: a biscuit token
+    /// :rtype: Biscuit
     pub fn build(&self, root: &PyPrivateKey) -> PyResult<PyBiscuit> {
         let keypair = KeyPair::from(&root.0);
         Ok(PyBiscuit(
@@ -81,39 +119,49 @@ impl PyBiscuitBuilder {
         ))
     }
 
-    /// Add code to the builder. This code snippet cannot contain parameters. See
-    /// `add_code_with_parameters` if you need to provide parameters
-    pub fn add_code(&mut self, source: &str) -> PyResult<()> {
-        self.0
-            .add_code(source)
-            .map_err(|e| DataLogError::new_err(e.to_string()))
-    }
-
     /// Add code to the builder, using the provided parameters.
-    pub fn add_code_with_parameters(
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
+    pub fn add_code(
         &mut self,
         source: &str,
-        raw_parameters: HashMap<String, PyTerm>,
-        scope_parameters: HashMap<String, PyPublicKey>,
+        parameters: Option<HashMap<String, PyTerm>>,
+        scope_parameters: Option<HashMap<String, PyPublicKey>>,
     ) -> PyResult<()> {
-        let mut parameters = HashMap::new();
+        let mut params = HashMap::new();
 
-        for (k, raw_value) in raw_parameters {
-            parameters.insert(k, raw_value.to_term()?);
+        if let Some(parameters) = parameters {
+            for (k, v) in parameters {
+                params.insert(k, v.to_term()?);
+            }
         }
 
-        let scope_parameters = scope_parameters
-            .iter()
-            .map(|(k, v)| (k.to_string(), v.0))
-            .collect();
+        let scope_params;
+
+        if let Some(scope_parameters) = scope_parameters {
+            scope_params = scope_parameters
+                .iter()
+                .map(|(k, v)| (k.to_string(), v.0))
+                .collect();
+        } else {
+            scope_params = HashMap::new();
+        }
 
         self.0
-            .add_code_with_params(source, parameters, scope_parameters)
+            .add_code_with_params(source, params, scope_params)
             .map_err(|e| DataLogError::new_err(e.to_string()))
     }
 
     /// Add a single fact to the builder. A single fact can be built with
     /// the `Fact` class and its constructor
+    ///
+    /// :param fact: a datalog fact
+    /// :type fact: Fact
     pub fn add_fact(&mut self, fact: &PyFact) -> PyResult<()> {
         self.0
             .add_fact(fact.0.clone())
@@ -122,6 +170,9 @@ impl PyBiscuitBuilder {
 
     /// Add a single rule to the builder. A single rule can be built with
     /// the `Rule` class and its constructor
+    ///
+    /// :param rule: a datalog rule
+    /// :type rule: Rule
     pub fn add_rule(&mut self, rule: &PyRule) -> PyResult<()> {
         self.0
             .add_rule(rule.0.clone())
@@ -130,17 +181,31 @@ impl PyBiscuitBuilder {
 
     /// Add a single check to the builder. A single check can be built with
     /// the `Check` class and its constructor
+    ///
+    /// :param check: a datalog check
+    /// :type check: Check
     pub fn add_check(&mut self, check: &PyCheck) -> PyResult<()> {
         self.0
             .add_check(check.0.clone())
             .map_err(|e| DataLogError::new_err(e.to_string()))
     }
 
-    /// Merge a `BlockBuilder` in this `BiscuitBuilder`. The `BlockBuilder` will not be modified
+    /// Merge a `BlockBuilder` in this `BiscuitBuilder`. The `BlockBuilder` parameter will not be modified
+    ///
+    /// :param builder: a datalog BlockBuilder
+    /// :type builder: BlockBuilder
     pub fn merge(&mut self, builder: &PyBlockBuilder) {
         self.0.merge(builder.0.clone())
     }
 
+    /// Set the root key identifier for this `BiscuitBuilder`
+    ///
+    /// :param root_key_id: the root key identifier
+    /// :type root_key_id: int
+    pub fn set_root_key_id(&mut self, root_key_id: u32) {
+        self.0.set_root_key_id(root_key_id)
+    }
+
     fn __repr__(&self) -> String {
         self.0.to_string()
     }
@@ -154,7 +219,8 @@ pub struct PyBiscuit(Biscuit);
 impl PyBiscuit {
     /// Creates a BiscuitBuilder
     ///
-    /// the builder can then create a new token with a root key
+    /// :return: an empty BiscuitBuilder
+    /// :rtype: BiscuitBuilder
     #[staticmethod]
     pub fn builder() -> PyResult<PyBiscuitBuilder> {
         PyBiscuitBuilder::new(None, None, None)
@@ -162,10 +228,17 @@ impl PyBiscuit {
 
     /// Deserializes a token from raw data
     ///
-    /// This will check the signature using the root key
+    /// This will check the signature using the provided root key (or function)
+    ///
+    /// :param data: raw biscuit bytes
+    /// :type data: bytes
+    /// :param root: either a public key or a function taking an integer (or `None`) and returning an public key
+    /// :type root: function,PublicKey
+    /// :return: the parsed and verified biscuit
+    /// :rtype: Biscuit
     #[classmethod]
-    pub fn from_bytes(_: &PyType, data: &[u8], root: &PyPublicKey) -> PyResult<PyBiscuit> {
-        match Biscuit::from(data, root.0) {
+    pub fn from_bytes(_: &PyType, data: &[u8], root: PyObject) -> PyResult<PyBiscuit> {
+        match Biscuit::from(data, PyKeyProvider { py_value: root }) {
             Ok(biscuit) => Ok(PyBiscuit(biscuit)),
             Err(error) => Err(BiscuitValidationError::new_err(error.to_string())),
         }
@@ -173,16 +246,26 @@ impl PyBiscuit {
 
     /// Deserializes a token from URL safe base 64 data
     ///
-    /// This will check the signature using the root key
+    /// This will check the signature using the provided root key (or function)
+    ///
+    /// :param data: a (url-safe) base64-encoded string
+    /// :type data: str
+    /// :param root: either a public key or a function taking an integer (or `None`) and returning an public key
+    /// :type root: function,PublicKey
+    /// :return: the parsed and verified biscuit
+    /// :rtype: Biscuit
     #[classmethod]
-    pub fn from_base64(_: &PyType, data: &str, root: &PyPublicKey) -> PyResult<PyBiscuit> {
-        match Biscuit::from_base64(data, root.0) {
+    pub fn from_base64(_: &PyType, data: &str, root: PyObject) -> PyResult<PyBiscuit> {
+        match Biscuit::from_base64(data, PyKeyProvider { py_value: root }) {
             Ok(biscuit) => Ok(PyBiscuit(biscuit)),
             Err(error) => Err(BiscuitValidationError::new_err(error.to_string())),
         }
     }
 
     /// Serializes to raw bytes
+    ///
+    /// :return: the serialized biscuit
+    /// :rtype: list
     pub fn to_bytes(&self) -> PyResult<Vec<u8>> {
         match self.0.to_vec() {
             Ok(vec) => Ok(vec),
@@ -191,16 +274,27 @@ impl PyBiscuit {
     }
 
     /// Serializes to URL safe base 64 data
+    ///
+    /// :return: the serialized biscuit
+    /// :rtype: str
     pub fn to_base64(&self) -> String {
         self.0.to_base64().unwrap()
     }
 
     /// Returns the number of blocks in the token
+    ///
+    /// :return: the number of blocks
+    /// :rtype: int
     pub fn block_count(&self) -> usize {
         self.0.block_count()
     }
 
     /// Prints a block's content as Datalog code
+    ///
+    /// :param index: the block index
+    /// :type index: int
+    /// :return: the code for the corresponding block
+    /// :rtype: str
     pub fn block_source(&self, index: usize) -> PyResult<String> {
         self.0
             .print_block_source(index)
@@ -208,6 +302,11 @@ impl PyBiscuit {
     }
 
     /// Create a new `Biscuit` by appending an attenuation block
+    ///
+    /// :param block: a builder for the new block
+    /// :type block: BlockBuilder
+    /// :return: the attenuated biscuit
+    /// :rtype: Biscuit
     pub fn append(&self, block: &PyBlockBuilder) -> PyResult<PyBiscuit> {
         self.0
             .append(block.0.clone())
@@ -221,12 +320,26 @@ impl PyBiscuit {
 }
 
 /// The Authorizer verifies a request according to its policies and the provided token
+///
+/// :param source: a datalog snippet
+/// :type source: str, optional
+/// :param parameters: values for the parameters in the datalog snippet
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "Authorizer")]
 pub struct PyAuthorizer(Authorizer);
 
 #[pymethods]
 impl PyAuthorizer {
     /// Create a new authorizer from a datalog snippet and optional parameter values
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
     #[new]
     pub fn new(
         source: Option<String>,
@@ -235,40 +348,54 @@ impl PyAuthorizer {
     ) -> PyResult<PyAuthorizer> {
         let mut builder = PyAuthorizer(Authorizer::new());
         if let Some(source) = source {
-            builder.add_code_with_parameters(
-                &source,
-                parameters.unwrap_or_default(),
-                scope_parameters.unwrap_or_default(),
-            )?;
+            builder.add_code(&source, parameters, scope_parameters)?;
         }
         Ok(builder)
     }
 
     /// Add code to the builder, using the provided parameters.
-    pub fn add_code_with_parameters(
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
+    pub fn add_code(
         &mut self,
         source: &str,
-        raw_parameters: HashMap<String, PyTerm>,
-        scope_parameters: HashMap<String, PyPublicKey>,
+        parameters: Option<HashMap<String, PyTerm>>,
+        scope_parameters: Option<HashMap<String, PyPublicKey>>,
     ) -> PyResult<()> {
-        let mut parameters = HashMap::new();
+        let mut params = HashMap::new();
 
-        for (k, raw_value) in raw_parameters {
-            parameters.insert(k, raw_value.to_term()?);
+        if let Some(parameters) = parameters {
+            for (k, v) in parameters {
+                params.insert(k, v.to_term()?);
+            }
         }
 
-        let scope_parameters = scope_parameters
-            .iter()
-            .map(|(k, v)| (k.to_string(), v.0))
-            .collect();
+        let scope_params;
+
+        if let Some(scope_parameters) = scope_parameters {
+            scope_params = scope_parameters
+                .iter()
+                .map(|(k, v)| (k.to_string(), v.0))
+                .collect();
+        } else {
+            scope_params = HashMap::new();
+        }
 
         self.0
-            .add_code_with_params(source, parameters, scope_parameters)
+            .add_code_with_params(source, params, scope_params)
             .map_err(|e| DataLogError::new_err(e.to_string()))
     }
 
     /// Add a single fact to the authorizer. A single fact can be built with
     /// the `Fact` class and its constructor
+    ///
+    /// :param fact: a datalog fact
+    /// :type fact: Fact
     pub fn add_fact(&mut self, fact: &PyFact) -> PyResult<()> {
         self.0
             .add_fact(fact.0.clone())
@@ -277,6 +404,9 @@ impl PyAuthorizer {
 
     /// Add a single rule to the authorizer. A single rule can be built with
     /// the `Rule` class and its constructor
+    ///
+    /// :param rule: a datalog rule
+    /// :type rule: Rule
     pub fn add_rule(&mut self, rule: &PyRule) -> PyResult<()> {
         self.0
             .add_rule(rule.0.clone())
@@ -285,6 +415,9 @@ impl PyAuthorizer {
 
     /// Add a single check to the authorizer. A single check can be built with
     /// the `Check` class and its constructor
+    ///
+    /// :param check: a datalog check
+    /// :type check: Check
     pub fn add_check(&mut self, check: &PyCheck) -> PyResult<()> {
         self.0
             .add_check(check.0.clone())
@@ -293,23 +426,35 @@ impl PyAuthorizer {
 
     /// Add a single policy to the authorizer. A single policy can be built with
     /// the `Policy` class and its constructor
+    ///
+    /// :param policy: a datalog policy
+    /// :type policy: Policy
     pub fn add_policy(&mut self, policy: &PyPolicy) -> PyResult<()> {
         self.0
             .add_policy(policy.0.clone())
             .map_err(|e| DataLogError::new_err(e.to_string()))
     }
 
-    /// Merge another `Authorizer` in this `Authorizer`. The `BlockBuilder` will not be modified
+    /// Merge another `Authorizer` in this `Authorizer`. The `Authorizer` argument will not be modified
+    ///
+    /// :param builder: an Authorizer
+    /// :type builder: Authorizer
     pub fn merge(&mut self, builder: &PyAuthorizer) {
         self.0.merge(builder.0.clone())
     }
 
     /// Merge a `BlockBuilder` in this `Authorizer`. The `BlockBuilder` will not be modified
+    ///
+    /// :param builder: a BlockBuilder
+    /// :type builder: BlockBuilder
     pub fn merge_block(&mut self, builder: &PyBlockBuilder) {
         self.0.merge_block(builder.0.clone())
     }
 
-    /// Add a Biscuit` to this `Authorizer`
+    /// Add a `Biscuit` to this `Authorizer`
+    ///
+    /// :param token: the token to authorize
+    /// :type token: Biscuit
     pub fn add_token(&mut self, token: &PyBiscuit) -> PyResult<()> {
         self.0
             .add_token(&token.0)
@@ -320,6 +465,9 @@ impl PyAuthorizer {
     ///
     /// Returns the index of the matching allow policy, or an error containing the matching deny
     /// policy or a list of the failing checks
+    ///
+    /// :return: the index of the matched allow rule
+    /// :rtype: int
     pub fn authorize(&mut self) -> PyResult<usize> {
         self.0
             .authorize()
@@ -331,6 +479,11 @@ impl PyAuthorizer {
     ///
     /// This function can be called before `authorize`, but in that case will only return facts that are directly defined,
     /// not the facts generated by rules.
+    ///
+    /// :param rule: a rule that will be ran against the authorizer contents
+    /// :type rule: Rule
+    /// :return: a list of generated facts
+    /// :rtype: list
     pub fn query(&mut self, rule: &PyRule) -> PyResult<Vec<PyFact>> {
         let results = self
             .0
@@ -349,12 +502,27 @@ impl PyAuthorizer {
 }
 
 /// Builder class allowing to create a block meant to be appended to an existing token
+///
+/// :param source: a datalog snippet
+/// :type source: str, optional
+/// :param parameters: values for the parameters in the datalog snippet
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "BlockBuilder")]
 #[derive(Clone)]
 pub struct PyBlockBuilder(builder::BlockBuilder);
 
 #[pymethods]
 impl PyBlockBuilder {
+    /// Create a builder from a datalog snippet and optional parameter values
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
     #[new]
     fn new(
         source: Option<String>,
@@ -363,17 +531,16 @@ impl PyBlockBuilder {
     ) -> PyResult<PyBlockBuilder> {
         let mut builder = PyBlockBuilder(builder::BlockBuilder::new());
         if let Some(source) = source {
-            builder.add_code_with_parameters(
-                &source,
-                parameters.unwrap_or_default(),
-                scope_parameters.unwrap_or_default(),
-            )?;
+            builder.add_code(&source, parameters, scope_parameters)?;
         }
         Ok(builder)
     }
 
     /// Add a single fact to the builder. A single fact can be built with
     /// the `Fact` class and its constructor
+    ///
+    /// :param fact: a datalog fact
+    /// :type fact: Fact
     pub fn add_fact(&mut self, fact: &PyFact) -> PyResult<()> {
         self.0
             .add_fact(fact.0.clone())
@@ -382,6 +549,9 @@ impl PyBlockBuilder {
 
     /// Add a single rule to the builder. A single rule can be built with
     /// the `Rule` class and its constructor
+    ///
+    /// :param rule: a datalog rule
+    /// :type rule: Rule
     pub fn add_rule(&mut self, rule: &PyRule) -> PyResult<()> {
         self.0
             .add_rule(rule.0.clone())
@@ -390,6 +560,9 @@ impl PyBlockBuilder {
 
     /// Add a single check to the builder. A single check can be built with
     /// the `Check` class and its constructor
+    ///
+    /// :param check: a datalog check
+    /// :type check: Check
     pub fn add_check(&mut self, check: &PyCheck) -> PyResult<()> {
         self.0
             .add_check(check.0.clone())
@@ -397,30 +570,48 @@ impl PyBlockBuilder {
     }
 
     /// Merge a `BlockBuilder` in this `BlockBuilder`. The `BlockBuilder` will not be modified
+    ///
+    /// :param builder: a datalog BlockBuilder
+    /// :type builder: BlockBuilder
     pub fn merge(&mut self, builder: &PyBlockBuilder) {
         self.0.merge(builder.0.clone())
     }
 
     /// Add code to the builder, using the provided parameters.
-    pub fn add_code_with_parameters(
+    ///
+    /// :param source: a datalog snippet
+    /// :type source: str, optional
+    /// :param parameters: values for the parameters in the datalog snippet
+    /// :type parameters: dict, optional
+    /// :param scope_parameters: public keys for the public key parameters in the datalog snippet
+    /// :type scope_parameters: dict, optional
+    pub fn add_code(
         &mut self,
         source: &str,
-        raw_parameters: HashMap<String, PyTerm>,
-        scope_parameters: HashMap<String, PyPublicKey>,
+        parameters: Option<HashMap<String, PyTerm>>,
+        scope_parameters: Option<HashMap<String, PyPublicKey>>,
     ) -> PyResult<()> {
-        let mut parameters = HashMap::new();
+        let mut params = HashMap::new();
 
-        for (k, raw_value) in raw_parameters {
-            parameters.insert(k, raw_value.to_term()?);
+        if let Some(parameters) = parameters {
+            for (k, v) in parameters {
+                params.insert(k, v.to_term()?);
+            }
         }
 
-        let scope_parameters = scope_parameters
-            .iter()
-            .map(|(k, v)| (k.to_string(), v.0))
-            .collect();
+        let scope_params;
+
+        if let Some(scope_parameters) = scope_parameters {
+            scope_params = scope_parameters
+                .iter()
+                .map(|(k, v)| (k.to_string(), v.0))
+                .collect();
+        } else {
+            scope_params = HashMap::new();
+        }
 
         self.0
-            .add_code_with_params(source, parameters, scope_parameters)
+            .add_code_with_params(source, params, scope_params)
             .map_err(|e| DataLogError::new_err(e.to_string()))
     }
 
@@ -442,6 +633,11 @@ impl PyKeyPair {
     }
 
     /// Generate a keypair from a private key
+    ///
+    /// :param private_key: the private key
+    /// :type private_key: PrivateKey
+    /// :return: the corresponding keypair
+    /// :rtype: KeyPair
     #[classmethod]
     pub fn from_private_key(_: &PyType, private_key: PyPrivateKey) -> Self {
         PyKeyPair(KeyPair::from(&private_key.0))
@@ -474,16 +670,27 @@ pub struct PyPublicKey(PublicKey);
 #[pymethods]
 impl PyPublicKey {
     /// Serializes a public key to raw bytes
+    ///
+    /// :return: the public key bytes
+    /// :rtype: list
     pub fn to_bytes(&self) -> [u8; 32] {
         self.0.to_bytes()
     }
 
     /// Serializes a public key to a hexadecimal string
+    ///
+    /// :return: the public key bytes (hex-encoded)
+    /// :rtype: str
     pub fn to_hex(&self) -> String {
         hex::encode(self.0.to_bytes())
     }
 
     /// Deserializes a public key from raw bytes
+    ///
+    /// :param data: the raw bytes
+    /// :type data: bytes
+    /// :return: the public key
+    /// :rtype: PublicKey
     #[classmethod]
     pub fn from_bytes(_: &PyType, data: &[u8]) -> PyResult<PyPublicKey> {
         match PublicKey::from_bytes(data) {
@@ -493,6 +700,11 @@ impl PyPublicKey {
     }
 
     /// Deserializes a public key from a hexadecimal string
+    ///
+    /// :param data: the hex-encoded string
+    /// :type data: str
+    /// :return: the public key
+    /// :rtype: PublicKey
     #[classmethod]
     pub fn from_hex(_: &PyType, data: &str) -> PyResult<PyPublicKey> {
         let data = match hex::decode(data) {
@@ -506,24 +718,35 @@ impl PyPublicKey {
     }
 }
 
+/// ed25519 private key
 #[pyclass(name = "PrivateKey")]
 #[derive(Clone)]
-/// ed25519 private key
 pub struct PyPrivateKey(PrivateKey);
 
 #[pymethods]
 impl PyPrivateKey {
-    /// Serializes a private key to raw bytes
+    /// Serializes a public key to raw bytes
+    ///
+    /// :return: the public key bytes
+    /// :rtype: list
     pub fn to_bytes(&self) -> [u8; 32] {
         self.0.to_bytes()
     }
 
     /// Serializes a private key to a hexadecimal string
+    ///
+    /// :return: the private key bytes (hex-encoded)
+    /// :rtype: str
     pub fn to_hex(&self) -> String {
         hex::encode(self.0.to_bytes())
     }
 
     /// Deserializes a private key from raw bytes
+    ///
+    /// :param data: the raw bytes
+    /// :type data: bytes
+    /// :return: the private key
+    /// :rtype: PrivateKey
     #[classmethod]
     pub fn from_bytes(_: &PyType, data: &[u8]) -> PyResult<PyPrivateKey> {
         match PrivateKey::from_bytes(data) {
@@ -533,6 +756,11 @@ impl PyPrivateKey {
     }
 
     /// Deserializes a private key from a hexadecimal string
+    ///
+    /// :param data: the hex-encoded string
+    /// :type data: str
+    /// :return: the private key
+    /// :rtype: PrivateKey
     #[classmethod]
     pub fn from_hex(_: &PyType, data: &str) -> PyResult<PyPrivateKey> {
         let data = match hex::decode(data) {
@@ -631,7 +859,12 @@ impl PyTerm {
     }
 }
 
-/// Datalog Fact
+/// A single datalog Fact
+///
+/// :param source: a datalog fact (without the ending semicolon)
+/// :type source: str
+/// :param parameters: values for the parameters in the datalog fact
+/// :type parameters: dict, optional
 #[pyclass(name = "Fact")]
 pub struct PyFact(builder::Fact);
 
@@ -686,6 +919,13 @@ impl PyFact {
 }
 
 /// A single datalog rule
+///
+/// :param source: a datalog rule (without the ending semicolon)
+/// :type source: str
+/// :param parameters: values for the parameters in the datalog rule
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog rule
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "Rule")]
 pub struct PyRule(builder::Rule);
 
@@ -723,6 +963,13 @@ impl PyRule {
 }
 
 /// A single datalog check
+///
+/// :param source: a datalog check (without the ending semicolon)
+/// :type source: str
+/// :param parameters: values for the parameters in the datalog check
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog check
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "Check")]
 pub struct PyCheck(builder::Check);
 
@@ -730,7 +977,11 @@ pub struct PyCheck(builder::Check);
 impl PyCheck {
     /// Build a check from the source and optional parameter values
     #[new]
-    pub fn new(source: &str, parameters: Option<HashMap<String, PyTerm>>) -> PyResult<Self> {
+    pub fn new(
+        source: &str,
+        parameters: Option<HashMap<String, PyTerm>>,
+        scope_parameters: Option<HashMap<String, PyPublicKey>>,
+    ) -> PyResult<Self> {
         let mut check: builder::Check = source
             .try_into()
             .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
@@ -741,6 +992,14 @@ impl PyCheck {
                     .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
             }
         }
+
+        if let Some(scope_parameters) = scope_parameters {
+            for (k, v) in scope_parameters {
+                check
+                    .set_scope(&k, v.0)
+                    .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
+            }
+        }
         Ok(PyCheck(check))
     }
 
@@ -750,6 +1009,13 @@ impl PyCheck {
 }
 
 /// A single datalog policy
+///
+/// :param source: a datalog policy (without the ending semicolon)
+/// :type source: str
+/// :param parameters: values for the parameters in the datalog policy
+/// :type parameters: dict, optional
+/// :param scope_parameters: public keys for the public key parameters in the datalog policy
+/// :type scope_parameters: dict, optional
 #[pyclass(name = "Policy")]
 pub struct PyPolicy(builder::Policy);
 
@@ -757,7 +1023,11 @@ pub struct PyPolicy(builder::Policy);
 impl PyPolicy {
     /// Build a check from the source and optional parameter values
     #[new]
-    pub fn new(source: &str, parameters: Option<HashMap<String, PyTerm>>) -> PyResult<Self> {
+    pub fn new(
+        source: &str,
+        parameters: Option<HashMap<String, PyTerm>>,
+        scope_parameters: Option<HashMap<String, PyPublicKey>>,
+    ) -> PyResult<Self> {
         let mut policy: builder::Policy = source
             .try_into()
             .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
@@ -768,6 +1038,14 @@ impl PyPolicy {
                     .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
             }
         }
+
+        if let Some(scope_parameters) = scope_parameters {
+            for (k, v) in scope_parameters {
+                policy
+                    .set_scope(&k, v.0)
+                    .map_err(|e: error::Token| DataLogError::new_err(e.to_string()))?;
+            }
+        }
         Ok(PyPolicy(policy))
     }
 
